@@ -44,9 +44,9 @@
 
 namespace {
 
-Resizable_buffer grbuf( initial_line_length );
+Resizable_buffer grbuf;
 bool archive_is_uncompressed_seekable = false;
-bool has_lz_ext;				// global var for archive_read
+bool archive_has_lz_ext;			// local var for archive_read
 
 bool skip_warn( const bool reset = false )	// avoid duplicate warnings
   {
@@ -91,8 +91,8 @@ bool make_path( const std::string & name )
 // Return value: 0 = OK, 1 = damaged member, 2 = fatal error.
 // If sizep and error, return in *sizep the number of bytes read.
 // The first 6 bytes of the archive must be intact for islz to be meaningful.
-int archive_read( const int infd, uint8_t * const buf, const int size,
-                  int * const sizep = 0 )
+int archive_read( const char * const archive_namep, const int infd,
+                  uint8_t * const buf, const int size, int * const sizep = 0 )
   {
   static LZ_Decoder * decoder = 0;
   static bool at_eof = false;
@@ -109,7 +109,8 @@ int archive_read( const int infd, uint8_t * const buf, const int size,
     const int rd = readblock( infd, buf, size );
     if( sizep ) *sizep = rd;
     if( rd != size && errno )
-      { show_error( "Error reading archive", errno ); fatal = true; return 2; }
+      { show_file_error( archive_namep, "Error reading archive", errno );
+        fatal = true; return 2; }
     const Lzip_header & header = (*(const Lzip_header *)buf);
     bool islz = ( rd >= min_member_size && header.verify_magic() &&
                   header.verify_version() &&
@@ -119,8 +120,8 @@ int archive_read( const int infd, uint8_t * const buf, const int size,
       ( !islz && !istar && rd == size && block_is_zero( buf, size ) );
     if( !islz && !istar && !iseof )		// corrupt or invalid format
       {
-      show_error( "This does not look like a POSIX tar archive." );
-      if( has_lz_ext && rd >= min_member_size ) islz = true;
+      show_file_error( archive_namep, posix_msg );
+      if( archive_has_lz_ext && rd >= min_member_size ) islz = true;
       if( !islz ) return 1;
       }
     if( !islz )						// uncompressed
@@ -132,10 +133,10 @@ int archive_read( const int infd, uint8_t * const buf, const int size,
         LZ_decompress_close( decoder ); fatal = true; return 2; }
     if( LZ_decompress_write( decoder, buf, rd ) != rd )
       internal_error( "library error (LZ_decompress_write)." );
-    const int res = archive_read( infd, buf, size, sizep );
+    const int res = archive_read( archive_namep, infd, buf, size, sizep );
     if( res != 0 ) { if( res == 2 ) fatal = true; return res; }
     if( verify_ustar_chksum( buf ) || block_is_zero( buf, size ) ) return 0;
-    show_error( "This does not look like a POSIX tar.lz archive." );
+    show_file_error( archive_namep, posix_lz_msg );
     fatal = true; return 2;
     }
 
@@ -143,7 +144,7 @@ int archive_read( const int infd, uint8_t * const buf, const int size,
     {
     const int rd = readblock( infd, buf, size ); if( rd == size ) return 0;
     if( sizep ) *sizep = rd;
-    show_error( "Archive ends unexpectedly." ); fatal = true; return 2;
+    show_file_error( archive_namep, end_msg ); fatal = true; return 2;
     }
   const int ibuf_size = 16384;
   uint8_t ibuf[ibuf_size];
@@ -159,11 +160,8 @@ int archive_read( const int infd, uint8_t * const buf, const int size,
       }
     if( rd == 0 && LZ_decompress_finished( decoder ) == 1 )
       { LZ_decompress_close( decoder );
-        show_error( "Archive ends unexpectedly." ); fatal = true; return 2; }
+        show_file_error( archive_namep, end_msg ); fatal = true; return 2; }
     sz += rd; if( sizep ) *sizep = sz;
-    if( sz == size && LZ_decompress_finished( decoder ) == 1 &&
-        LZ_decompress_close( decoder ) < 0 )
-      { show_error( "LZ_decompress_close failed." ); fatal = true; return 2; }
     if( sz < size && !at_eof && LZ_decompress_write_size( decoder ) > 0 )
       {
       const int rsize = std::min( ibuf_size, LZ_decompress_write_size( decoder ) );
@@ -174,8 +172,8 @@ int archive_read( const int infd, uint8_t * const buf, const int size,
         {
         at_eof = true; LZ_decompress_finish( decoder );
         if( errno )
-          { show_error( "Error reading archive", errno ); fatal = true;
-            return 2; }
+          { show_file_error( archive_namep, "Error reading archive", errno );
+            fatal = true; return 2; }
         }
       }
     }
@@ -247,7 +245,7 @@ bool block_is_zero( const uint8_t * const buf, const int size )
   }
 
 
-void format_member_name( const Extended & extended, const Tar_header header,
+bool format_member_name( const Extended & extended, const Tar_header header,
                          Resizable_buffer & rbuf, const bool long_format )
   {
   if( long_format )
@@ -279,45 +277,50 @@ void format_member_name( const Extended & extended, const Tar_header header,
                 1900 + tm->tm_year, 1 + tm->tm_mon, tm->tm_mday,
                 tm->tm_hour, tm->tm_min, extended.path().c_str(),
                 link_string, islink ? extended.linkpath().c_str() : "" );
-      if( (int)rbuf.size() > len + offset || !rbuf.resize( len + offset + 1 ) )
-        break;
+      if( (int)rbuf.size() > len + offset ) break;
+      if( !rbuf.resize( len + offset + 1 ) ) return false;
       }
     }
   else
     {
-    if( rbuf.size() < extended.path().size() + 2 )
-      rbuf.resize( extended.path().size() + 2 );
+    if( rbuf.size() < extended.path().size() + 2 &&
+        !rbuf.resize( extended.path().size() + 2 ) ) return false;
     snprintf( rbuf(), rbuf.size(), "%s\n", extended.path().c_str() );
     }
+  return true;
+  }
+
+
+bool show_member_name( const Extended & extended, const Tar_header header,
+                       const int vlevel, Resizable_buffer & rbuf )
+  {
+  if( verbosity >= vlevel )
+    {
+    if( !format_member_name( extended, header, rbuf, verbosity > vlevel ) )
+      { show_error( mem_msg ); return false; }
+    std::fputs( rbuf(), stdout );
+    std::fflush( stdout );
+    }
+  return true;
   }
 
 namespace {
 
-void show_member_name( const Extended & extended, const Tar_header header,
-                       const int vlevel, Resizable_buffer & rbuf )
+int skip_member( const char * const archive_namep, const int infd,
+                 const Extended & extended )
   {
-  if( verbosity < vlevel ) return;
-  format_member_name( extended, header, rbuf, verbosity > vlevel );
-  std::fputs( rbuf(), stdout );
-  std::fflush( stdout );
-  }
-
-
-int skip_member( const int infd, const Extended & extended )
-  {
-  unsigned long long rest = extended.file_size();
+  long long rest = extended.file_size();
   const int rem = rest % header_size;
-  const int padding = rem ? header_size - rem : 0;
-  if( archive_is_uncompressed_seekable &&
-      lseek( infd, rest + padding, SEEK_CUR ) > 0 ) return 0;
-  const unsigned bufsize = 32 * header_size;
+  if( rem ) rest += header_size - rem;		// padding
+  if( archive_is_uncompressed_seekable && lseek( infd, rest, SEEK_CUR ) > 0 )
+    return 0;
+  const int bufsize = 32 * header_size;
   uint8_t buf[bufsize];
   while( rest > 0 )
     {
-    const int rsize = ( rest >= bufsize ) ? bufsize : rest + padding;
-    const int ret = archive_read( infd, buf, rsize );
+    const int rsize = ( rest >= bufsize ) ? bufsize : rest;
+    const int ret = archive_read( archive_namep, infd, buf, rsize );
     if( ret != 0 ) { if( ret == 2 ) return 2; else break; }
-    if( rest < bufsize ) break;
     rest -= rsize;
     }
   return 0;
@@ -326,20 +329,22 @@ int skip_member( const int infd, const Extended & extended )
 
 void show_file_diff( const char * const filename, const char * const msg )
   {
-  if( verbosity >= 0 ) std::fprintf( stderr, "%s: %s\n", filename, msg );
+  if( verbosity >= 0 )
+    { std::printf( "%s: %s\n", filename, msg ); std::fflush( stdout ); }
   }
 
 
-int compare_member( const int infd1, const Extended & extended,
-                    const Tar_header header, const bool ignore_ids )
+int compare_member( const char * const archive_namep, const int infd1,
+                    const Extended & extended, const Tar_header header,
+                    const bool ignore_ids )
   {
-  show_member_name( extended, header, 1, grbuf );
-  unsigned long long rest = extended.file_size();
+  if( !show_member_name( extended, header, 1, grbuf ) ) return 1;
+  long long rest = extended.file_size();
   const char * const filename = extended.path().c_str();
   const Typeflag typeflag = (Typeflag)header[typeflag_o];
   bool diff = false, size_differs = false, type_differs = true;
   struct stat st;
-  if( lstat( filename, &st ) != 0 )
+  if( hstat( filename, &st ) != 0 )
     show_file_error( filename, "Warning: Can't stat", errno );
   else if( ( typeflag == tf_regular || typeflag == tf_hiperf ) &&
            !S_ISREG( st.st_mode ) )
@@ -380,7 +385,7 @@ int compare_member( const int infd1, const Extended & extended,
           { show_file_diff( filename, "Mod time differs" ); diff = true; }
         }
       if( ( typeflag == tf_regular || typeflag == tf_hiperf ) &&
-          (off_t)rest != st.st_size )		// don't compare contents
+          rest != st.st_size )			// don't compare contents
         { show_file_diff( filename, "Size differs" ); size_differs = true; }
       if( ( typeflag == tf_chardev || typeflag == tf_blockdev ) &&
           ( parse_octal( header + devmajor_o, devmajor_l ) !=
@@ -406,24 +411,26 @@ int compare_member( const int infd1, const Extended & extended,
     }
   if( diff || size_differs || type_differs )
     { diff = false; set_error_status( 1 ); }
-  if( rest == 0 ) return 0;
+  if( rest <= 0 ) return 0;
   if( ( typeflag != tf_regular && typeflag != tf_hiperf ) ||
-      size_differs || type_differs ) return skip_member( infd1, extended );
+      size_differs || type_differs )
+    return skip_member( archive_namep, infd1, extended );
   // else compare file contents
   const int rem = rest % header_size;
   const int padding = rem ? header_size - rem : 0;
-  const unsigned bufsize = 32 * header_size;
+  const int bufsize = 32 * header_size;
   uint8_t buf1[bufsize];
   uint8_t buf2[bufsize];
   const int infd2 = open_instream( filename );
   if( infd2 < 0 )
-    { set_error_status( 1 ); return skip_member( infd1, extended ); }
+    { set_error_status( 1 );
+      return skip_member( archive_namep, infd1, extended ); }
   int retval = 0;
   while( rest > 0 )
     {
     const int rsize1 = ( rest >= bufsize ) ? bufsize : rest + padding;
     const int rsize2 = ( rest >= bufsize ) ? bufsize : rest;
-    const int ret = archive_read( infd1, buf1, rsize1 );
+    const int ret = archive_read( archive_namep, infd1, buf1, rsize1 );
     if( ret != 0 ) { if( ret == 2 ) retval = 2; diff = true; break; }
     if( !diff )
       {
@@ -450,11 +457,11 @@ int compare_member( const int infd1, const Extended & extended,
   }
 
 
-int list_member( const int infd, const Extended & extended,
-                 const Tar_header header )
+int list_member( const char * const archive_namep, const int infd,
+                 const Extended & extended, const Tar_header header )
   {
-  show_member_name( extended, header, 0, grbuf );
-  return skip_member( infd, extended );
+  if( !show_member_name( extended, header, 0, grbuf ) ) return 1;
+  return skip_member( archive_namep, infd, extended );
   }
 
 
@@ -466,14 +473,15 @@ bool contains_dotdot( const char * const filename )
   }
 
 
-int extract_member( const int infd, const Extended & extended,
-                    const Tar_header header, const bool keep_damaged )
+int extract_member( const char * const archive_namep, const int infd,
+                    const Extended & extended, const Tar_header header,
+                    const bool keep_damaged )
   {
   const char * const filename = extended.path().c_str();
   if( contains_dotdot( filename ) )
     {
     show_file_error( filename, "Contains a '..' component, skipping." );
-    return skip_member( infd, extended );
+    return skip_member( archive_namep, infd, extended );
     }
   const mode_t mode = parse_octal( header + mode_o, mode_l );	 // 12 bits
   const time_t mtime = parse_octal( header + mtime_o, mtime_l ); // 33 bits
@@ -481,7 +489,7 @@ int extract_member( const int infd, const Extended & extended,
   const bool islink = ( typeflag == tf_link || typeflag == tf_symlink );
   int outfd = -1;
 
-  show_member_name( extended, header, 1, grbuf );
+  if( !show_member_name( extended, header, 1, grbuf ) ) return 1;
   std::remove( filename );
   make_path( filename );
   switch( typeflag )
@@ -551,22 +559,22 @@ int extract_member( const int infd, const Extended & extended,
     return 2;
     }
 
-  const unsigned bufsize = 32 * header_size;
+  const int bufsize = 32 * header_size;
   uint8_t buf[bufsize];
-  unsigned long long rest = extended.file_size();
+  long long rest = extended.file_size();
   const int rem = rest % header_size;
   const int padding = rem ? header_size - rem : 0;
   while( rest > 0 )
     {
     const int rsize = ( rest >= bufsize ) ? bufsize : rest + padding;
     int rd;
-    const int ret = archive_read( infd, buf, rsize, &rd );
+    const int ret = archive_read( archive_namep, infd, buf, rsize, &rd );
     if( ret != 0 )
       {
       if( outfd >= 0 )
         {
         if( keep_damaged )
-          { writeblock( outfd, buf, std::min( rest, (unsigned long long)rd ) );
+          { writeblock( outfd, buf, std::min( rest, (long long)rd ) );
             close( outfd ); }
         else { close( outfd ); std::remove( filename ); }
         }
@@ -614,18 +622,17 @@ bool compare_tslash( const char * const name1, const char * const name2 )
 
 namespace {
 
-bool parse_records( const int infd, Extended & extended,
-                    const Tar_header header, const bool permissive )
+bool parse_records( const char * const archive_namep, const int infd,
+                    Extended & extended, const Tar_header header,
+                    Resizable_buffer & rbuf, const bool permissive )
   {
-  const unsigned long long edsize = parse_octal( header + size_o, size_l );
-  const unsigned long long bufsize = round_up( edsize );
-  if( bufsize == 0 || edsize == 0 || edsize >= 1ULL << 33 )
+  const long long edsize = parse_octal( header + size_o, size_l );
+  const long long bufsize = round_up( edsize );
+  if( edsize <= 0 || edsize >= 1LL << 33 || bufsize >= INT_MAX )
     return false;				// overflow or no extended data
-  char * const buf = new char[bufsize];		// extended records buffer
-  const bool ret = ( archive_read( infd, (uint8_t *)buf, bufsize ) == 0 &&
-                     extended.parse( buf, edsize, permissive ) );
-  delete[] buf;
-  return ret;
+  if( !rbuf.resize( bufsize ) ) return false;	// extended records buffer
+  return ( archive_read( archive_namep, infd, (uint8_t *)rbuf(), bufsize ) == 0 &&
+           extended.parse( rbuf(), edsize, permissive ) );
   }
 
 } // end namespace
@@ -685,8 +692,10 @@ int decode( const std::string & archive_name, const Arg_parser & parser,
             const bool keep_damaged, const bool missing_crc,
             const bool permissive )
   {
-  const int infd = archive_name.size() ?
-                   open_instream( archive_name ) : STDIN_FILENO;
+  const bool from_stdin = archive_name.empty();
+  const char * const archive_namep =
+    from_stdin ? "(stdin)" : archive_name.c_str();
+  const int infd = from_stdin ? STDIN_FILENO : open_instream( archive_name );
   if( infd < 0 ) return 1;
 
   // Execute -C options and mark filenames to be compared, extracted or listed.
@@ -702,7 +711,9 @@ int decode( const std::string & archive_name, const Arg_parser & parser,
         { show_file_error( dir, "Error changing working directory", errno );
           return 1; }
       }
-    if( !code && parser.argument( i ).size() ) name_pending[i] = true;
+    if( !code && parser.argument( i ).size() &&
+        !Exclude::excluded( parser.argument( i ).c_str() ) )
+      name_pending[i] = true;
     }
 
   // multi-threaded --list is faster even with 1 thread and 1 file in archive
@@ -712,9 +723,10 @@ int decode( const std::string & archive_name, const Arg_parser & parser,
     const long members = lzip_index.members();
     if( lzip_index.retval() == 0 && members >= 2 )	// one file + eof
       {
-      // show_file_error( archive_name.c_str(), "Is compressed seekable" );
-      return list_lz( parser, name_pending, lzip_index, filenames, debug_level,
-                      infd, std::min( (long)num_workers, members ),
+      // show_file_error( archive_namep, "Is compressed seekable" );
+      return list_lz( archive_namep, parser, name_pending, lzip_index,
+                      filenames, debug_level, infd,
+                      std::min( (long)num_workers, members ),
                       missing_crc, permissive );
       }
     if( lseek( infd, 0, SEEK_SET ) == 0 && lzip_index.retval() != 0 &&
@@ -722,50 +734,52 @@ int decode( const std::string & archive_name, const Arg_parser & parser,
       archive_is_uncompressed_seekable = true;	// unless compressed corrupt
     }
 
-  has_lz_ext =				// global var for archive_read
-    ( archive_name.size() > 3 &&
-      archive_name.compare( archive_name.size() - 3, 3, ".lz" ) == 0 ) ||
-    ( archive_name.size() > 4 &&
-      archive_name.compare( archive_name.size() - 4, 4, ".tlz" ) == 0 );
+  archive_has_lz_ext = has_lz_ext( archive_name );	// var for archive_read
   Extended extended;			// metadata from extended records
   int retval = 0;
   bool prev_extended = false;		// prev header was extended
-  while( true )				// process one tar member per iteration
+  while( true )				// process one tar header per iteration
     {
     Tar_header header;
-    const int ret = archive_read( infd, header, header_size );
-    if( ret == 2 ) return 2;
+    const int ret = archive_read( archive_namep, infd, header, header_size );
+    if( ret == 2 ) { retval = 2; break; }
     if( ret != 0 || !verify_ustar_chksum( header ) )
       {
-      if( ret == 0 && block_is_zero( header, header_size ) ) break;	// EOF
+      if( ret == 0 && block_is_zero( header, header_size ) )
+        {
+        if( !prev_extended || permissive ) break;		// EOF
+        show_file_error( archive_namep, fv_msg1 );
+        retval = 2; break;
+        }
       if( skip_warn() && verbosity >= 2 )
         std::fprintf( stderr, "ustar chksum = %07o\n", ustar_chksum( header ) );
       set_error_status( 2 ); continue;
       }
-    skip_warn( true );			// reset warning
+    skip_warn( true );				// reset warning
 
     const Typeflag typeflag = (Typeflag)header[typeflag_o];
     if( typeflag == tf_global )
       {
-      if( prev_extended )
-        { show_error( "Format violation: global header after extended header." );
-          return 2; }
+      if( prev_extended && !permissive )
+        { show_file_error( archive_namep, fv_msg2 ); retval = 2; break; }
       Extended dummy;		// global headers are parsed and ignored
-      if( !parse_records( infd, dummy, header, true ) )
-        { show_error( "Error in global extended records. Skipping to next header." );
+      if( !parse_records( archive_namep, infd, dummy, header, grbuf, true ) )
+        { show_file_error( archive_namep,
+            "Error in global extended records. Skipping to next header." );
           set_error_status( 2 ); }
       continue;
       }
     if( typeflag == tf_extended )
       {
       if( prev_extended && !permissive )
-        { show_error( "Format violation: consecutive extended headers found."
-                      /*" Use --permissive.", 0, true*/ ); return 2; }
-      if( !parse_records( infd, extended, header, permissive ) )
-        { show_error( "Error in extended records. Skipping to next header." );
+        { show_file_error( archive_namep, fv_msg3 ); retval = 2; break; }
+      if( !parse_records( archive_namep, infd, extended, header, grbuf,
+                          permissive ) )
+        { show_file_error( archive_namep,
+            "Error in extended records. Skipping to next header." );
           extended.reset(); set_error_status( 2 ); }
       else if( !extended.crc_present() && missing_crc )
-        { show_error( "Missing CRC in extended records.", 0, true ); return 2; }
+        { show_file_error( archive_namep, mcrc_msg ); retval = 2; break; }
       prev_extended = true;
       continue;
       }
@@ -773,23 +787,25 @@ int decode( const std::string & archive_name, const Arg_parser & parser,
 
     extended.fill_from_ustar( header );		// copy metadata from header
 
-    const bool skip = check_skip_filename( parser, name_pending,
-                                       extended.path().c_str(), filenames );
-    if( skip )
-      retval = skip_member( infd, extended );
+    if( check_skip_filename( parser, name_pending, extended.path().c_str(),
+                             filenames ) )
+      retval = skip_member( archive_namep, infd, extended );
     else if( program_mode == m_list )
-      retval = list_member( infd, extended, header );
+      retval = list_member( archive_namep, infd, extended, header );
     else if( program_mode == m_diff )
-      retval = compare_member( infd, extended, header, ignore_ids );
-    else
-      retval = extract_member( infd, extended, header, keep_damaged );
+      retval = compare_member( archive_namep, infd, extended, header, ignore_ids );
+    else retval = extract_member( archive_namep, infd, extended, header,
+                                  keep_damaged );
     extended.reset();
     if( retval )
-      { show_error( "Error is not recoverable: exiting now." );
-        return retval; }
+      { show_error( "Error is not recoverable: exiting now." ); break; }
     }
 
-  for( int i = 0; i < parser.arguments(); ++i )
+  if( close( infd ) != 0 && !retval )
+    { show_file_error( archive_namep, "Error closing archive", errno );
+      retval = 1; }
+
+  if( retval == 0 ) for( int i = 0; i < parser.arguments(); ++i )
     if( !parser.code( i ) && parser.argument( i ).size() && name_pending[i] )
       {
       show_file_error( parser.argument( i ).c_str(), "Not found in archive." );
