@@ -69,7 +69,7 @@ void Lzip_index::set_num_error( const char * const msg, unsigned long long num )
 
 
 // If successful, push last member and set pos to member header.
-bool Lzip_index::skip_trailing_data( const int fd, long long & pos,
+bool Lzip_index::skip_trailing_data( const int fd, unsigned long long & pos,
                    const bool ignore_trailing, const bool loose_trailing )
   {
   enum { block_size = 16384,
@@ -151,7 +151,7 @@ Lzip_index::Lzip_index( const int infd, const bool ignore_trailing,
   if( !isvalid_ds( header.dictionary_size() ) )
     { error_ = bad_dict_msg; retval_ = 2; return; }
 
-  long long pos = insize;	// always points to a header or to EOF
+  unsigned long long pos = insize;	// always points to a header or to EOF
   while( pos >= min_member_size )
     {
     Lzip_trailer trailer;
@@ -159,7 +159,7 @@ Lzip_index::Lzip_index( const int infd, const bool ignore_trailing,
                    pos - Lzip_trailer::size ) != Lzip_trailer::size )
       { set_errno_error( "Error reading member trailer: " ); break; }
     const unsigned long long member_size = trailer.member_size();
-    if( member_size > (unsigned long long)pos || !trailer.verify_consistency() )
+    if( member_size > pos || !trailer.verify_consistency() )
       {
       if( member_vector.empty() )
         { if( skip_trailing_data( infd, pos, ignore_trailing, loose_trailing ) )
