@@ -36,6 +36,7 @@ inbad1="${testdir}"/test_bad1.txt
 inbad2="${testdir}"/test_bad2.txt
 test3="${testdir}"/test3.tar
 test3_lz="${testdir}"/test3.tar.lz
+test3dir="${testdir}"/test3_dir.tar
 test3dir_lz="${testdir}"/test3_dir.tar.lz
 test3dot_lz="${testdir}"/test3_dot.tar.lz
 tarint1_lz="${testdir}"/tar_in_tlz1.tar.lz
@@ -55,6 +56,7 @@ bad3_lz="${testdir}"/test3_bad3.tar.lz
 bad4_lz="${testdir}"/test3_bad4.tar.lz
 bad5_lz="${testdir}"/test3_bad5.tar.lz
 bad6_lz="${testdir}"/test3_bad6.tar.lz
+eof="${testdir}"/eof.tar
 eof_lz="${testdir}"/eof.tar.lz
 fail=0
 lwarn=0
@@ -66,13 +68,14 @@ lzlib_1_11() { [ ${lwarn} = 0 ] &&
 # Description of test files for tarlz:
 # test.txt.tar.lz:   1 member (test.txt).
 # t155.tar[.lz]:     directory + links + file + eof, all with 155 char names
+# t155_fv?.tar[.lz]: like t155.tar but with 3 kinds of format violations
 # tar_in_tlz1.tar.lz 2 members (test.txt.tar test3.tar) 3 lzip members
 # tar_in_tlz2.tar.lz 2 members (test.txt.tar test3.tar) 5 lzip members
 # ts_in_link.tar.lz: 4 symbolic links (link[1-4]) to / /dir/ dir/ dir(107/)
 # test_bad1.tar.lz:  truncated at offset 6000 (of 7495)
 # test_bad2.tar.lz:  byte at offset 6000 changed from 0x56 to 0x46
-# test3.tar:         3 members (foo bar baz) + 2 zeroed 512-byte blocks
-# test3_dir.tar.lz:  like test3.tar.lz but members /dir/foo /dir/bar /dir/baz
+# test3.tar[.lz]:    3 members (foo bar baz) + 2 zeroed 512-byte blocks
+# test3_dir.tar[.lz] like test3.tar but members /dir/foo /dir/bar /dir/baz
 # test3_dot.tar.lz:  3 times 3 members ./foo ././bar ./././baz
 #                    the 3 central members with filename in extended header
 # test3_bad1.tar:    byte at offset  259 changed from 't' to '0' (magic)
@@ -87,11 +90,19 @@ lzlib_1_11() { [ ${lwarn} = 0 ] &&
 # test3_bad4.tar.lz: combined damage of test3_bad2.tar.lz and test3_bad3.tar.lz
 # test3_bad5.tar.lz: [71-134] --> zeroed (first trailer + seconf header)
 # test3_bad6.tar.lz: 510 zeros prepended to test3.tar.lz (header in two blocks)
+# test3_eof?.tar:    like test3_eof?.tar.lz but uncompressed
 # test3_eof1.tar.lz: test3.tar.lz without eof blocks
 # test3_eof2.tar.lz: test3.tar.lz with only one eof block
 # test3_eof3.tar.lz: test3.tar.lz with one zeroed block between foo and bar
+# test3_eof4.tar.lz: test3.tar.lz ended by extended header without eof blocks
+# test3_eof5.tar.lz: test3.tar.lz split ext first member, without eof blocks
 # test3_em?.tar.lz:  test3.tar.lz with one empty lzip member at each position
 # test3_em6.tar.lz:  test3.tar.lz preceded by four empty lzip members
+# test3_gh?.tar:     test3.tar with global header at each position
+# test3_gh?.tar.lz:  test3.tar.lz with global before bar split in 4 ways
+# test3_gh5.tar.lz:  test3.tar.lz with global in lzip member before foo
+# test3_gh6.tar.lz:  test3.tar.lz with global before foo in same member
+# test3_sm?.tar.lz:  test3.tar.lz with extended bar member split in 4 ways
 # tlz_in_tar1.tar:   1 member (test3.tar.lz) first magic damaged
 # tlz_in_tar2.tar:   2 members (foo test3.tar.lz) first magic damaged
 # ug32chars.tar.lz:  1 member (foo) with 32-character owner and group names
@@ -112,17 +123,13 @@ printf "testing tarlz-%s..." "$2"
 [ ! -e out.tar.lz ] || test_failed $LINENO
 "${TARLZ}" -rf out.tar.lz || test_failed $LINENO
 [ ! -e out.tar.lz ] || test_failed $LINENO
-"${TARLZ}" -q -rf - "${in}"
-[ $? = 1 ] || test_failed $LINENO
-[ ! -e - ] || test_failed $LINENO
-"${TARLZ}" -q -r "${in}"
-[ $? = 1 ] || test_failed $LINENO
+"${TARLZ}" -r || test_failed $LINENO
 "${TARLZ}" --uncompressed -q -rf out.tar "${in}"
 [ $? = 1 ] || test_failed $LINENO
 [ ! -e out.tar ] || test_failed $LINENO
 cat "${test3_lz}" > test.tar.lz || framework_failure
 "${TARLZ}" --uncompressed -q -rf test.tar.lz "${in}"
-[ $? = 1 ] || test_failed $LINENO
+[ $? = 2 ] || test_failed $LINENO
 cmp "${test3_lz}" test.tar.lz || test_failed $LINENO
 rm -f test.tar.lz || framework_failure
 cat "${test3}" > test.tar || framework_failure
@@ -175,7 +182,16 @@ rm -f test.txt || framework_failure
 cmp "${in}" test.txt || test_failed $LINENO
 rm -f test.txt || framework_failure
 
-# reference files for cmp
+# test3 reference files for diff
+"${TARLZ}" -tf "${test3}" > list3 || test_failed $LINENO
+"${TARLZ}" -tvf "${test3}" > vlist3 || test_failed $LINENO
+"${TARLZ}" -tf "${test3_lz}" > out || test_failed $LINENO
+diff -u list3 out || test_failed $LINENO
+"${TARLZ}" -tvf "${test3_lz}" > out || test_failed $LINENO
+diff -u vlist3 out || test_failed $LINENO
+rm -f out || framework_failure
+
+# test3 reference files for cmp
 cat "${testdir}"/rfoo > cfoo || framework_failure
 cat "${testdir}"/rbar > cbar || framework_failure
 cat "${testdir}"/rbaz > cbaz || framework_failure
@@ -185,7 +201,10 @@ cmp cfoo foo || test_failed $LINENO
 cmp cbar bar || test_failed $LINENO
 cmp cbaz baz || test_failed $LINENO
 rm -f foo bar baz || framework_failure
-"${TARLZ}" -q -tf "${test3_lz}" ./foo ./bar ./baz || test_failed $LINENO
+"${TARLZ}" -tvf "${test3_lz}" ./foo ./bar ./baz > out 2> /dev/null ||
+	test_failed $LINENO
+diff -u vlist3 out || test_failed $LINENO
+rm -f out || framework_failure
 "${TARLZ}" -q -xf "${test3_lz}" ./foo ./bar ./baz || test_failed $LINENO
 cmp cfoo foo || test_failed $LINENO
 cmp cbar bar || test_failed $LINENO
@@ -212,67 +231,153 @@ cmp cfoo foo || test_failed $LINENO
 cmp cbar bar || test_failed $LINENO
 cmp cbaz baz || test_failed $LINENO
 rm -f foo bar baz || framework_failure
-"${TARLZ}" -q -xf "${test3dir_lz}" --missing-crc || test_failed $LINENO
+
+for i in "${test3dir}" "${test3dir_lz}" ; do
+	"${TARLZ}" -q -tf "$i" --missing-crc || test_failed $LINENO "$i"
+	"${TARLZ}" -q -xf "$i" --missing-crc || test_failed $LINENO "$i"
+	cmp cfoo dir/foo || test_failed $LINENO "$i"
+	cmp cbar dir/bar || test_failed $LINENO "$i"
+	cmp cbaz dir/baz || test_failed $LINENO "$i"
+	rm -rf dir || framework_failure
+	"${TARLZ}" -q -tf "$i" dir || test_failed $LINENO "$i"
+	"${TARLZ}" -q -xf "$i" dir || test_failed $LINENO "$i"
+	cmp cfoo dir/foo || test_failed $LINENO "$i"
+	cmp cbar dir/bar || test_failed $LINENO "$i"
+	cmp cbaz dir/baz || test_failed $LINENO "$i"
+	rm -rf dir || framework_failure
+	"${TARLZ}" -q -tf "$i" dir/foo dir/baz || test_failed $LINENO "$i"
+	"${TARLZ}" -q -xf "$i" dir/foo dir/baz || test_failed $LINENO "$i"
+	cmp cfoo dir/foo || test_failed $LINENO "$i"
+	[ ! -e dir/bar ] || test_failed $LINENO "$i"
+	cmp cbaz dir/baz || test_failed $LINENO "$i"
+	rm -rf dir || framework_failure
+done
+
+# --exclude
+"${TARLZ}" -xf "${test3}" --exclude='f*o' --exclude=baz || test_failed $LINENO
+[ ! -e foo ] || test_failed $LINENO
+cmp cbar bar || test_failed $LINENO
+[ ! -e baz ] || test_failed $LINENO
+rm -f foo bar baz || framework_failure
+"${TARLZ}" -xf "${test3_lz}" --exclude=bar || test_failed $LINENO
+cmp cfoo foo || test_failed $LINENO
+[ ! -e bar ] || test_failed $LINENO
+cmp cbaz baz || test_failed $LINENO
+rm -f foo bar baz || framework_failure
+"${TARLZ}" -q -xf "${test3dir_lz}" --exclude='?ar' || test_failed $LINENO
 cmp cfoo dir/foo || test_failed $LINENO
-cmp cbar dir/bar || test_failed $LINENO
+[ ! -e dir/bar ] || test_failed $LINENO
 cmp cbaz dir/baz || test_failed $LINENO
 rm -rf dir || framework_failure
-"${TARLZ}" -q -tf "${test3dir_lz}" dir/foo dir/bar dir/baz || test_failed $LINENO
-"${TARLZ}" -q -xf "${test3dir_lz}" dir/foo dir/bar dir/baz || test_failed $LINENO
+"${TARLZ}" -q -xf "${test3dir_lz}" --exclude=dir/bar || test_failed $LINENO
 cmp cfoo dir/foo || test_failed $LINENO
-cmp cbar dir/bar || test_failed $LINENO
+[ ! -e dir/bar ] || test_failed $LINENO
 cmp cbaz dir/baz || test_failed $LINENO
 rm -rf dir || framework_failure
-#
-"${TARLZ}" -q -tf "${testdir}"/test3_eof1.tar.lz
+"${TARLZ}" -q -xf "${test3dir_lz}" --exclude=dir || test_failed $LINENO
+[ ! -e dir ] || test_failed $LINENO
+rm -rf dir || framework_failure
+"${TARLZ}" -q -xf "${test3dir_lz}" --exclude='dir/*' || test_failed $LINENO
+[ ! -e dir ] || test_failed $LINENO
+rm -rf dir || framework_failure
+"${TARLZ}" -q -xf "${test3dir_lz}" --exclude='[bf][ao][orz]' ||
+	test_failed $LINENO
+[ ! -e dir ] || test_failed $LINENO
+rm -rf dir || framework_failure
+"${TARLZ}" -q -xf "${test3dir_lz}" --exclude='*o' dir/foo || test_failed $LINENO
+[ ! -e dir ] || test_failed $LINENO
+rm -rf dir || framework_failure
+
+# eof
+"${TARLZ}" -tvf "${testdir}"/test3_eof1.tar > out 2> /dev/null
 [ $? = 2 ] || test_failed $LINENO
-"${TARLZ}" -q -tf "${testdir}"/test3_eof2.tar.lz || test_failed $LINENO
-"${TARLZ}" -q -tf "${testdir}"/test3_eof3.tar.lz || test_failed $LINENO
-"${TARLZ}" -q -n0 -tf "${testdir}"/test3_eof1.tar.lz
+diff -u vlist3 out || test_failed $LINENO
+"${TARLZ}" -tvf "${testdir}"/test3_eof2.tar > out || test_failed $LINENO
+diff -u vlist3 out || test_failed $LINENO
+"${TARLZ}" -q -tf "${testdir}"/test3_eof3.tar || test_failed $LINENO
+"${TARLZ}" -tvf "${testdir}"/test3_eof4.tar > out 2> /dev/null
 [ $? = 2 ] || test_failed $LINENO
-"${TARLZ}" -q -n0 -tf "${testdir}"/test3_eof2.tar.lz || test_failed $LINENO
-"${TARLZ}" -q -n0 -tf "${testdir}"/test3_eof3.tar.lz || test_failed $LINENO
+diff -u vlist3 out || test_failed $LINENO
+for i in 0 2 6 ; do
+	"${TARLZ}" -n$i -tvf "${testdir}"/test3_eof1.tar.lz > out 2> /dev/null
+	[ $? = 2 ] || test_failed $LINENO $i
+	diff -u vlist3 out || test_failed $LINENO
+	"${TARLZ}" -n$i -tvf "${testdir}"/test3_eof2.tar.lz > out ||
+		test_failed $LINENO $i
+	diff -u vlist3 out || test_failed $LINENO
+	"${TARLZ}" -q -n$i -tf "${testdir}"/test3_eof3.tar.lz ||
+		test_failed $LINENO $i
+	"${TARLZ}" -n$i -tvf "${testdir}"/test3_eof4.tar.lz > out 2> /dev/null
+	[ $? = 2 ] || test_failed $LINENO $i
+	diff -u vlist3 out || test_failed $LINENO
+	"${TARLZ}" -n$i -tvf "${testdir}"/test3_eof5.tar.lz > out 2> /dev/null
+	[ $? = 2 ] || test_failed $LINENO $i
+	diff -u vlist3 out || test_failed $LINENO
+done
+rm -f out || framework_failure
 #
-"${TARLZ}" -q -xf "${testdir}"/test3_eof1.tar.lz
+"${TARLZ}" -q -xf "${testdir}"/test3_eof1.tar
 [ $? = 2 ] || test_failed $LINENO
 cmp cfoo foo || test_failed $LINENO
 cmp cbar bar || test_failed $LINENO
 cmp cbaz baz || test_failed $LINENO
 rm -f foo bar baz || framework_failure
-"${TARLZ}" -xf "${testdir}"/test3_eof2.tar.lz || test_failed $LINENO
+"${TARLZ}" -xf "${testdir}"/test3_eof2.tar || test_failed $LINENO
 cmp cfoo foo || test_failed $LINENO
 cmp cbar bar || test_failed $LINENO
 cmp cbaz baz || test_failed $LINENO
 rm -f foo bar baz || framework_failure
-"${TARLZ}" -xf "${testdir}"/test3_eof3.tar.lz || test_failed $LINENO
+"${TARLZ}" -xf "${testdir}"/test3_eof3.tar || test_failed $LINENO
 cmp cfoo foo || test_failed $LINENO
 [ ! -e bar ] || test_failed $LINENO
 [ ! -e baz ] || test_failed $LINENO
 rm -f foo bar baz || framework_failure
-#
-"${TARLZ}" -q -n0 -xf "${testdir}"/test3_eof1.tar.lz
+"${TARLZ}" -q -xf "${testdir}"/test3_eof4.tar
 [ $? = 2 ] || test_failed $LINENO
 cmp cfoo foo || test_failed $LINENO
 cmp cbar bar || test_failed $LINENO
 cmp cbaz baz || test_failed $LINENO
 rm -f foo bar baz || framework_failure
-"${TARLZ}" -n0 -xf "${testdir}"/test3_eof2.tar.lz || test_failed $LINENO
-cmp cfoo foo || test_failed $LINENO
-cmp cbar bar || test_failed $LINENO
-cmp cbaz baz || test_failed $LINENO
-rm -f foo bar baz || framework_failure
-"${TARLZ}" -n0 -xf "${testdir}"/test3_eof3.tar.lz || test_failed $LINENO
-cmp cfoo foo || test_failed $LINENO
-[ ! -e bar ] || test_failed $LINENO
-[ ! -e baz ] || test_failed $LINENO
-rm -f foo bar baz || framework_failure
+#
+for i in 0 2 6 ; do
+	"${TARLZ}" -q -n$i -xf "${testdir}"/test3_eof1.tar.lz
+	[ $? = 2 ] || test_failed $LINENO $i
+	cmp cfoo foo || test_failed $LINENO $i
+	cmp cbar bar || test_failed $LINENO $i
+	cmp cbaz baz || test_failed $LINENO $i
+	rm -f foo bar baz || framework_failure
+	"${TARLZ}" -n$i -xf "${testdir}"/test3_eof2.tar.lz ||
+		test_failed $LINENO $i
+	cmp cfoo foo || test_failed $LINENO $i
+	cmp cbar bar || test_failed $LINENO $i
+	cmp cbaz baz || test_failed $LINENO $i
+	rm -f foo bar baz || framework_failure
+	"${TARLZ}" -n$i -xf "${testdir}"/test3_eof3.tar.lz ||
+		test_failed $LINENO $i
+	cmp cfoo foo || test_failed $LINENO $i
+	[ ! -e bar ] || test_failed $LINENO $i
+	[ ! -e baz ] || test_failed $LINENO $i
+	rm -f foo bar baz || framework_failure
+	"${TARLZ}" -q -n$i -xf "${testdir}"/test3_eof4.tar.lz
+	[ $? = 2 ] || test_failed $LINENO $i
+	cmp cfoo foo || test_failed $LINENO $i
+	cmp cbar bar || test_failed $LINENO $i
+	cmp cbaz baz || test_failed $LINENO $i
+	rm -f foo bar baz || framework_failure
+	"${TARLZ}" -q -n$i -xf "${testdir}"/test3_eof5.tar.lz
+	[ $? = 2 ] || test_failed $LINENO $i
+	cmp cfoo foo || test_failed $LINENO $i
+	cmp cbar bar || test_failed $LINENO $i
+	cmp cbaz baz || test_failed $LINENO $i
+	rm -f foo bar baz || framework_failure
+done
 
 # test --list and --extract tar in tar.lz
 for i in "${tarint1_lz}" "${tarint2_lz}" ; do
 	for j in 0 2 6 ; do
-		"${TARLZ}" -tf "$i" --threads=$j > out$j ||
+		"${TARLZ}" -tf "$i" -n$j > out$j ||
 			test_failed $LINENO "$i $j"
-		"${TARLZ}" -tvf "$i" --threads=$j > outv$j ||
+		"${TARLZ}" -tvf "$i" -n$j > outv$j ||
 			test_failed $LINENO "$i $j"
 	done
 	diff -u out0 out2 || test_failed $LINENO $i
@@ -288,23 +393,35 @@ for i in "${tarint1_lz}" "${tarint2_lz}" ; do
 	rm -f test.txt.tar test3.tar || framework_failure
 done
 
-# test --list and --extract with empty lzip members
-for i in 1 2 3 4 5 6 ; do
+# test --list and --extract with global headers uncompressed
+for i in gh1 gh2 gh3 gh4 ; do
+	"${TARLZ}" -tf "${testdir}"/test3_${i}.tar > out ||
+		test_failed $LINENO $i
+	diff -u list3 out || test_failed $LINENO $i
+	"${TARLZ}" -tvf "${testdir}"/test3_${i}.tar > out ||
+		test_failed $LINENO $i
+	diff -u vlist3 out || test_failed $LINENO $i
+	"${TARLZ}" -xf "${testdir}"/test3_${i}.tar || test_failed $LINENO $i
+	cmp cfoo foo || test_failed $LINENO $i
+	cmp cbar bar || test_failed $LINENO $i
+	cmp cbaz baz || test_failed $LINENO $i
+	rm -f foo bar baz out || framework_failure
+done
+
+# test --list and --extract with empty lzip members, global headers and
+# extended tar members split among lzip members
+for i in em1 em2 em3 em4 em5 em6 gh1 gh2 gh3 gh4 gh5 gh6 sm1 sm2 sm3 sm4 ; do
 	for j in 0 2 6 ; do
-		"${TARLZ}" -tf "${testdir}"/test3_em${i}.tar.lz --threads=$j \
-			> out$j || test_failed $LINENO "$i $j"
-		"${TARLZ}" -tvf "${testdir}"/test3_em${i}.tar.lz --threads=$j \
-			> outv$j || test_failed $LINENO "$i $j"
+		"${TARLZ}" -tf "${testdir}"/test3_${i}.tar.lz -n$j > out ||
+			test_failed $LINENO "$i $j"
+		diff -u list3 out || test_failed $LINENO "$i $j"
+		"${TARLZ}" -tvf "${testdir}"/test3_${i}.tar.lz -n$j > out ||
+			test_failed $LINENO "$i $j"
+		diff -u vlist3 out || test_failed $LINENO "$i $j"
 	done
-	diff -u out0 out2 || test_failed $LINENO $i
-	diff -u out0 out6 || test_failed $LINENO $i
-	diff -u out2 out6 || test_failed $LINENO $i
-	diff -u outv0 outv2 || test_failed $LINENO $i
-	diff -u outv0 outv6 || test_failed $LINENO $i
-	diff -u outv2 outv6 || test_failed $LINENO $i
-	rm -f out0 out2 out6 outv0 outv2 outv6 || framework_failure
+	rm -f out || framework_failure
 	for j in 0 2 6 ; do
-		"${TARLZ}" -xf "${testdir}"/test3_em${i}.tar.lz --threads=$j ||
+		"${TARLZ}" -xf "${testdir}"/test3_${i}.tar.lz -n$j ||
 			test_failed $LINENO "$i $j"
 		cmp cfoo foo || test_failed $LINENO "$i $j"
 		cmp cbar bar || test_failed $LINENO "$i $j"
@@ -313,7 +430,10 @@ for i in 1 2 3 4 5 6 ; do
 	done
 done
 
-# test --concatenate
+# test --concatenate compressed
+cat "${in}" > out.tar.lz || framework_failure		# invalid tar.lz
+"${TARLZ}" -Aqf out.tar.lz "${test3_lz}"
+[ $? = 2 ] || test_failed $LINENO
 cat "${in_tar_lz}" > out.tar.lz || framework_failure
 "${TARLZ}" -Af out.tar.lz "${test3_lz}" || test_failed $LINENO
 "${TARLZ}" -xf out.tar.lz || test_failed $LINENO
@@ -321,14 +441,82 @@ cmp "${in}" test.txt || test_failed $LINENO
 cmp cfoo foo || test_failed $LINENO
 cmp cbar bar || test_failed $LINENO
 cmp cbaz baz || test_failed $LINENO
+rm -f test.txt foo bar baz || framework_failure
 touch aout.tar.lz || framework_failure		# concatenate to empty file
+"${TARLZ}" -Aqf aout.tar.lz "${in_tar}"
+[ $? = 2 ] || test_failed $LINENO
+"${TARLZ}" -Af aout.tar.lz "${in_tar_lz}" "${test3_lz}" || test_failed $LINENO
+cmp out.tar.lz aout.tar.lz || test_failed $LINENO
+"${TARLZ}" -Af aout.tar.lz || test_failed $LINENO	# concatenate nothing
+cmp out.tar.lz aout.tar.lz || test_failed $LINENO
+"${TARLZ}" -Aqf aout.tar.lz aout.tar.lz || test_failed $LINENO
+cmp out.tar.lz aout.tar.lz || test_failed $LINENO
+"${TARLZ}" -Aq "${in_tar_lz}" "${test3}" > aout.tar.lz		# to stdout
+[ $? = 2 ] || test_failed $LINENO
+cmp "${in_tar_lz}" aout.tar.lz || test_failed $LINENO
+"${TARLZ}" -A "${in_tar_lz}" "${test3_lz}" > aout.tar.lz || test_failed $LINENO
+cmp out.tar.lz aout.tar.lz || test_failed $LINENO
+cat "${eof_lz}" > aout.tar.lz || framework_failure  # concatenate to empty archive
+"${TARLZ}" -Aqf aout.tar.lz "${in_tar}"
+[ $? = 2 ] || test_failed $LINENO
 "${TARLZ}" -Af aout.tar.lz "${in_tar_lz}" "${test3_lz}" || test_failed $LINENO
 cmp out.tar.lz aout.tar.lz || test_failed $LINENO
 cat "${in_tar_lz}" > aout.tar.lz || framework_failure
 "${TARLZ}" -Aqf aout.tar.lz "${test3_lz}" "${test3}"
 [ $? = 2 ] || test_failed $LINENO
 cmp out.tar.lz aout.tar.lz || test_failed $LINENO
-rm -f test.txt foo bar baz out.tar.lz aout.tar.lz || framework_failure
+rm -f aout.tar.lz || framework_failure
+touch aout.tar.lz || framework_failure			# --exclude
+"${TARLZ}" -Af aout.tar.lz "${in_tar_lz}" "${test3_lz}" --exclude 'test3*' ||
+	test_failed $LINENO
+"${TARLZ}" -Af aout.tar.lz "${in_tar_lz}" "${test3_lz}" --exclude '*txt*' ||
+	test_failed $LINENO
+cmp out.tar.lz aout.tar.lz || test_failed $LINENO
+rm -f out.tar.lz aout.tar.lz || framework_failure
+
+# test --concatenate uncompressed
+cat "${in}" > out.tar || framework_failure		# invalid tar
+"${TARLZ}" -Aqf out.tar "${test3}"
+[ $? = 2 ] || test_failed $LINENO
+cat "${in_tar}" > out.tar || framework_failure
+"${TARLZ}" -Af out.tar "${test3}" || test_failed $LINENO
+"${TARLZ}" -xf out.tar || test_failed $LINENO
+cmp "${in}" test.txt || test_failed $LINENO
+cmp cfoo foo || test_failed $LINENO
+cmp cbar bar || test_failed $LINENO
+cmp cbaz baz || test_failed $LINENO
+rm -f test.txt foo bar baz || framework_failure
+touch aout.tar || framework_failure		# concatenate to empty file
+"${TARLZ}" -Aqf aout.tar "${in_tar_lz}"
+[ $? = 2 ] || test_failed $LINENO
+"${TARLZ}" -Af aout.tar "${in_tar}" "${test3}" || test_failed $LINENO
+cmp out.tar aout.tar || test_failed $LINENO
+"${TARLZ}" -Af aout.tar || test_failed $LINENO	# concatenate nothing
+cmp out.tar aout.tar || test_failed $LINENO
+"${TARLZ}" -Aqf aout.tar aout.tar || test_failed $LINENO
+cmp out.tar aout.tar || test_failed $LINENO
+"${TARLZ}" -Aq "${in_tar}" "${test3_lz}" > aout.tar		# to stdout
+[ $? = 2 ] || test_failed $LINENO
+cmp "${in_tar}" aout.tar || test_failed $LINENO
+"${TARLZ}" -A "${in_tar}" "${test3}" > aout.tar || test_failed $LINENO
+cmp out.tar aout.tar || test_failed $LINENO
+cat "${eof}" > aout.tar || framework_failure	# concatenate to empty archive
+"${TARLZ}" -Aqf aout.tar "${in_tar_lz}"
+[ $? = 2 ] || test_failed $LINENO
+"${TARLZ}" -Af aout.tar "${in_tar}" "${test3}" || test_failed $LINENO
+cmp out.tar aout.tar || test_failed $LINENO
+cat "${in_tar}" > aout.tar || framework_failure
+"${TARLZ}" -Aqf aout.tar "${test3}" "${test3_lz}"
+[ $? = 2 ] || test_failed $LINENO
+cmp out.tar aout.tar || test_failed $LINENO
+rm -f aout.tar || framework_failure
+touch aout.tar || framework_failure			# --exclude
+"${TARLZ}" -Af aout.tar "${test3}" "${in_tar}" --exclude 'test3*' ||
+	test_failed $LINENO
+"${TARLZ}" -Af aout.tar "${test3}" "${in_tar}" --exclude '*txt*' ||
+	test_failed $LINENO
+cmp out.tar aout.tar || test_failed $LINENO
+rm -f out.tar aout.tar || framework_failure
 
 # test --create
 cat "${in}" > test.txt || framework_failure
@@ -346,7 +534,7 @@ rm -f test.txt out.tar out.tar.lz || framework_failure
 cat cfoo > foo || framework_failure
 rm -f bar || framework_failure
 cat cbaz > baz || framework_failure
-"${TARLZ}" -q -cf out.tar.lz foo bar baz
+"${TARLZ}" -0 -q -cf out.tar.lz foo bar baz
 [ $? = 1 ] || test_failed $LINENO
 rm -f foo bar baz || framework_failure
 "${TARLZ}" -xf out.tar.lz --missing-crc || test_failed $LINENO
@@ -364,13 +552,9 @@ rm -f out.tar.lz || framework_failure
 cat cfoo > foo || framework_failure
 cat cbar > bar || framework_failure
 cat cbaz > baz || framework_failure
-"${TARLZ}" -0 -cf out.tar.lz foo bar baz || test_failed $LINENO
+"${TARLZ}" -0 -cf out.tar.lz foo bar baz --out-slots=1 || test_failed $LINENO
 "${TARLZ}" -0 -q -cf aout.tar.lz foo bar aout.tar.lz baz || test_failed $LINENO
 cmp out.tar.lz aout.tar.lz || test_failed $LINENO
-"${TARLZ}" -q -Af aout.tar.lz aout.tar.lz || test_failed $LINENO
-cmp out.tar.lz aout.tar.lz || test_failed $LINENO
-"${TARLZ}" -q -rf aout.tar.lz aout.tar.lz || test_failed $LINENO
-cmp out.tar.lz aout.tar.lz || test_failed $LINENO
 rm -f aout.tar.lz || framework_failure
 #
 "${TARLZ}" -0 -cf aout.tar.lz foo bar baz -C / || test_failed $LINENO
@@ -425,16 +609,173 @@ cmp cbaz dir1/baz || test_failed $LINENO
 rm -rf dir1 || framework_failure
 rm -f out.tar.lz aout.tar.lz || framework_failure
 
+# --exclude
+cat cfoo > foo || framework_failure
+cat cbar > bar || framework_failure
+cat cbaz > baz || framework_failure
+"${TARLZ}" -0 -cf out.tar.lz foo bar baz --exclude 'ba?' || test_failed $LINENO
+rm -f foo bar baz || framework_failure
+"${TARLZ}" -xf out.tar.lz || test_failed $LINENO
+cmp cfoo foo || test_failed $LINENO
+[ ! -e bar ] || test_failed $LINENO
+[ ! -e baz ] || test_failed $LINENO
+rm -f out.tar.lz foo bar baz || framework_failure
+cat cfoo > foo || framework_failure
+cat cbar > bar || framework_failure
+cat cbaz > baz || framework_failure
+"${TARLZ}" --un -cf out.tar foo bar baz --exclude 'ba*' || test_failed $LINENO
+rm -f foo bar baz || framework_failure
+"${TARLZ}" -xf out.tar || test_failed $LINENO
+cmp cfoo foo || test_failed $LINENO
+[ ! -e bar ] || test_failed $LINENO
+[ ! -e baz ] || test_failed $LINENO
+rm -f out.tar foo bar baz || framework_failure
+
+# test --delete
+for e in "" .lz ; do
+  "${TARLZ}" -A "${in_tar}"$e "${test3}"$e > out.tar$e || test_failed $LINENO $e
+  "${TARLZ}" -f out.tar$e --delete test.txt || test_failed $LINENO $e
+  cmp "${test3}"$e out.tar$e || test_failed $LINENO $e
+  "${TARLZ}" -f out.tar$e --delete || test_failed $LINENO $e	# delete nothing
+  cmp "${test3}"$e out.tar$e || test_failed $LINENO $e
+  "${TARLZ}" -qf out.tar$e --delete nx_file
+  [ $? = 1 ] || test_failed $LINENO $e
+  cmp "${test3}"$e out.tar$e || test_failed $LINENO $e
+  "${TARLZ}" -A "${in_tar}"$e "${test3dir}"$e > out.tar$e || test_failed $LINENO $e
+  "${TARLZ}" -qf out.tar$e --delete test.txt || test_failed $LINENO $e
+  cmp "${test3dir}"$e out.tar$e || test_failed $LINENO $e
+  "${TARLZ}" -A "${in_tar}"$e "${test3dir}"$e > out.tar$e || test_failed $LINENO $e
+  "${TARLZ}" -qf out.tar$e --delete dir || test_failed $LINENO $e
+  cmp "${in_tar}"$e out.tar$e || test_failed $LINENO $e
+  "${TARLZ}" -A "${in_tar}"$e "${test3dir}"$e > out.tar$e || test_failed $LINENO $e
+  "${TARLZ}" -qf out.tar$e --del dir/foo dir/bar dir/baz || test_failed $LINENO $e
+  cmp "${in_tar}"$e out.tar$e || test_failed $LINENO $e
+  "${TARLZ}" -A "${in_tar}"$e "${test3dir}"$e > out.tar$e || test_failed $LINENO $e
+  "${TARLZ}" -qf out.tar$e --del dir/foo dir/baz || test_failed $LINENO $e
+  cmp "${in_tar}"$e out.tar$e > /dev/null && test_failed $LINENO $e
+  "${TARLZ}" -qf out.tar$e --del dir/bar || test_failed $LINENO $e
+  cmp "${in_tar}"$e out.tar$e || test_failed $LINENO $e
+  "${TARLZ}" -A "${in_tar}"$e "${test3}"$e > out.tar$e || test_failed $LINENO $e
+  "${TARLZ}" -f out.tar$e --delete foo bar baz || test_failed $LINENO $e
+  cmp "${in_tar}"$e out.tar$e || test_failed $LINENO $e
+  "${TARLZ}" -A "${in_tar}"$e "${test3}"$e > out.tar$e || test_failed $LINENO $e
+  "${TARLZ}" -f out.tar$e --del test.txt foo bar baz || test_failed $LINENO $e
+  cmp "${eof}"$e out.tar$e || test_failed $LINENO $e
+  "${TARLZ}" -A "${in_tar}"$e "${test3}"$e > out.tar$e || test_failed $LINENO $e
+  for i in test.txt foo bar baz ; do
+    "${TARLZ}" -f out.tar$e --delete $i || test_failed $LINENO "$e $i"
+  done
+  cmp "${eof}"$e out.tar$e || test_failed $LINENO $e
+  "${TARLZ}" -A "${in_tar}"$e "${test3}"$e > out.tar$e || test_failed $LINENO $e
+  for i in baz bar foo test.txt ; do
+    "${TARLZ}" -f out.tar$e --delete $i || test_failed $LINENO "$e $i"
+  done
+  cmp "${eof}"$e out.tar$e || test_failed $LINENO $e
+  "${TARLZ}" -A "${in_tar}"$e "${test3}"$e > out.tar$e || test_failed $LINENO $e
+  for i in foo bar test.txt baz ; do
+    "${TARLZ}" -f out.tar$e --delete $i || test_failed $LINENO "$e $i"
+  done
+  cmp "${eof}"$e out.tar$e || test_failed $LINENO $e
+  "${TARLZ}" -A "${in_tar}"$e "${t155}"$e "${test3}"$e > out.tar$e ||
+    test_failed $LINENO $e
+  "${TARLZ}" -f out.tar$e --del baz foo test.txt bar || test_failed $LINENO $e
+  cmp "${t155}"$e out.tar$e || test_failed $LINENO $e
+  "${TARLZ}" -f out.tar$e --delete link || test_failed $LINENO $e
+  "${TARLZ}" -q -tf out.tar$e || test_failed $LINENO
+  cmp "${t155}"$e out.tar$e > /dev/null && test_failed $LINENO $e
+  rm -f out.tar$e || framework_failure
+done
+
+# test --delete individual member after collective member
+cat cfoo > foo || framework_failure
+cat cbar > bar || framework_failure
+cat cbaz > baz || framework_failure
+cat "${in}" > test.txt || framework_failure
+"${TARLZ}" -0 -cf out.tar.lz foo bar baz --asolid || test_failed $LINENO
+"${TARLZ}" -0 -rf out.tar.lz test.txt || test_failed $LINENO
+rm -f foo bar baz test.txt || framework_failure
+"${TARLZ}" -f out.tar.lz --delete test.txt || test_failed $LINENO
+"${TARLZ}" -xf out.tar.lz || test_failed $LINENO
+cmp cfoo foo || test_failed $LINENO
+cmp cbar bar || test_failed $LINENO
+cmp cbaz baz || test_failed $LINENO
+[ ! -e test.txt ] || test_failed $LINENO
+rm -f out.tar.lz foo bar baz test.txt || framework_failure
+
+# test --delete with empty lzip member, global header
+for i in 1 2 3 4 5 6 ; do
+	cat "${testdir}"/test3_em${i}.tar.lz > out.tar.lz || framework_failure
+	for j in foo bar baz ; do
+		"${TARLZ}" -f out.tar.lz --delete $j || test_failed $LINENO "$i $j"
+	done
+	rm -f out.tar.lz || framework_failure
+done
+cat "${testdir}"/test3_gh5.tar.lz > out.tar.lz || framework_failure
+for i in foo bar baz ; do
+	"${TARLZ}" -f out.tar.lz --delete $i || test_failed $LINENO $i
+done
+rm -f out.tar.lz || framework_failure
+for i in 1 2 3 4 ; do
+	cat "${testdir}"/test3_gh${i}.tar > out.tar || framework_failure
+	for j in foo bar baz ; do
+		"${TARLZ}" -f out.tar --delete $j || test_failed $LINENO "$i $j"
+	done
+	rm -f out.tar || framework_failure
+done
+
+# test --dereference
+touch dummy_file || framework_failure
+if ln dummy_file dummy_link 2> /dev/null &&
+   ln -s dummy_file dummy_slink 2> /dev/null ; then
+	ln_works=yes
+else
+	printf "\nwarning: skipping link test: 'ln' does not work on your system."
+fi
+rm -f dummy_slink dummy_link dummy_file || framework_failure
+
+if [ "${ln_works}" = yes ] ; then
+	mkdir dir || framework_failure
+	cat cfoo > dir/foo || framework_failure
+	cat cbar > dir/bar || framework_failure
+	cat cbaz > dir/baz || framework_failure
+	ln -s dir dir_link || framework_failure
+	"${TARLZ}" -0 -cf out1 dir_link || test_failed $LINENO
+	"${TARLZ}" --un -cf out2 dir_link || test_failed $LINENO
+	"${TARLZ}" -0 -n0 -cf out3 dir_link || test_failed $LINENO
+	"${TARLZ}" -0 -h -cf hout1 dir_link || test_failed $LINENO
+	"${TARLZ}" --un -h -cf hout2 dir_link || test_failed $LINENO
+	"${TARLZ}" -0 -n0 -h -cf hout3 dir_link || test_failed $LINENO
+	rm -rf dir dir_link || framework_failure
+	for i in 1 2 3 ; do
+		"${TARLZ}" -xf out$i || test_failed $LINENO $i
+		[ -h dir_link ] || test_failed $LINENO $i
+		[ ! -e dir_link/foo ] || test_failed $LINENO $i
+		[ ! -e dir_link/bar ] || test_failed $LINENO $i
+		[ ! -e dir_link/baz ] || test_failed $LINENO $i
+		rm -rf dir_link out$i || framework_failure
+		"${TARLZ}" -xf hout$i || test_failed $LINENO $i
+		[ -d dir_link ] || test_failed $LINENO $i
+		cmp cfoo dir_link/foo || test_failed $LINENO $i
+		cmp cbar dir_link/bar || test_failed $LINENO $i
+		cmp cbaz dir_link/baz || test_failed $LINENO $i
+		rm -rf dir_link hout$i || framework_failure
+	done
+fi
+
 # test --append
 cat cfoo > foo || framework_failure
 cat cbar > bar || framework_failure
 cat cbaz > baz || framework_failure
-"${TARLZ}" -0 -cf out.tar.lz foo bar baz || test_failed $LINENO
+"${TARLZ}" -0 -cf out.tar.lz foo bar baz --out-slots=1024 || test_failed $LINENO
 "${TARLZ}" -0 -cf nout.tar.lz foo bar baz --no-solid || test_failed $LINENO
 "${TARLZ}" -0 -cf aout.tar.lz foo || test_failed $LINENO
 "${TARLZ}" -0 -rf aout.tar.lz bar baz --no-solid || test_failed $LINENO
 cmp nout.tar.lz aout.tar.lz || test_failed $LINENO
 rm -f nout.tar.lz aout.tar.lz || framework_failure
+touch aout.tar || framework_failure		# wrong extension empty file
+"${TARLZ}" -0 -rf aout.tar foo bar baz || test_failed $LINENO
+cmp out.tar.lz aout.tar || test_failed $LINENO
+rm -f aout.tar || framework_failure
 touch aout.tar.lz || framework_failure		# append to empty file
 "${TARLZ}" -0 -rf aout.tar.lz foo bar baz || test_failed $LINENO
 cmp out.tar.lz aout.tar.lz || test_failed $LINENO
@@ -445,11 +786,57 @@ cmp out.tar.lz aout.tar.lz || test_failed $LINENO
 "${TARLZ}" -0 -q -rf aout.tar.lz nx_file
 [ $? = 1 ] || test_failed $LINENO
 cmp out.tar.lz aout.tar.lz || test_failed $LINENO
-cat "${eof_lz}" > aout.tar.lz || framework_failure	# append to empty archive
+"${TARLZ}" -0 -q -rf aout.tar.lz aout.tar.lz || test_failed $LINENO
+cmp out.tar.lz aout.tar.lz || test_failed $LINENO
+"${TARLZ}" -0 -r foo bar baz > aout.tar.lz || test_failed $LINENO  # to stdout
+cmp out.tar.lz aout.tar.lz || test_failed $LINENO
+"${TARLZ}" --un -q -rf aout.tar.lz foo bar baz	# wrong extension archive
+[ $? = 2 ] || test_failed $LINENO
+cmp out.tar.lz aout.tar.lz || test_failed $LINENO
+cat "${eof_lz}" > aout.tar.lz || framework_failure  # append to empty archive
 "${TARLZ}" -0 -rf aout.tar.lz foo bar baz || test_failed $LINENO
 cmp out.tar.lz aout.tar.lz || test_failed $LINENO
+"${TARLZ}" --un -q -rf aout.tar.lz foo bar baz	# wrong extension empty archive
+[ $? = 2 ] || test_failed $LINENO
+cmp out.tar.lz aout.tar.lz || test_failed $LINENO
 rm -f out.tar.lz aout.tar.lz || framework_failure
 
+# --uncompressed
+"${TARLZ}" --un -cf out.tar foo bar baz || test_failed $LINENO
+"${TARLZ}" --un -cf aout.tar foo || test_failed $LINENO
+"${TARLZ}" --un -rf aout.tar foo bar baz --exclude foo || test_failed $LINENO
+cmp out.tar aout.tar || test_failed $LINENO
+rm -f aout.tar || framework_failure
+touch aout.tar.lz empty || framework_failure	# wrong extension empty file
+"${TARLZ}" --un -q -rf aout.tar.lz foo bar baz
+[ $? = 2 ] || test_failed $LINENO
+cmp aout.tar.lz empty || test_failed $LINENO
+rm -f aout.tar.lz empty || framework_failure
+touch aout.tar || framework_failure		# append to empty file
+"${TARLZ}" --un -rf aout.tar foo bar baz || test_failed $LINENO
+cmp out.tar aout.tar || test_failed $LINENO
+"${TARLZ}" --un -rf aout.tar || test_failed $LINENO	# append nothing
+cmp out.tar aout.tar || test_failed $LINENO
+"${TARLZ}" --un -rf aout.tar -C nx_dir || test_failed $LINENO
+cmp out.tar aout.tar || test_failed $LINENO
+"${TARLZ}" --un -q -rf aout.tar nx_file
+[ $? = 1 ] || test_failed $LINENO
+cmp out.tar aout.tar || test_failed $LINENO
+"${TARLZ}" --un -q -rf aout.tar aout.tar || test_failed $LINENO
+cmp out.tar aout.tar || test_failed $LINENO
+"${TARLZ}" --un -r foo bar baz > aout.tar || test_failed $LINENO  # to stdout
+cmp out.tar aout.tar || test_failed $LINENO
+"${TARLZ}" -0 -q -rf aout.tar foo bar baz	# wrong extension archive
+[ $? = 2 ] || test_failed $LINENO
+cmp out.tar aout.tar || test_failed $LINENO
+cat "${eof}" > aout.tar || framework_failure	# append to empty archive
+"${TARLZ}" --un -rf aout.tar foo bar baz || test_failed $LINENO
+cmp out.tar aout.tar || test_failed $LINENO
+"${TARLZ}" -0 -q -rf aout.tar foo bar baz	# wrong extension empty archive
+[ $? = 2 ] || test_failed $LINENO
+cmp out.tar aout.tar || test_failed $LINENO
+rm -f out.tar aout.tar || framework_failure
+
 # append to solid archive
 "${TARLZ}" --solid -q -0 -cf out.tar.lz "${in}" foo bar || test_failed $LINENO
 "${TARLZ}" -q -tf out.tar.lz || test_failed $LINENO	# compressed seekable
@@ -483,9 +870,20 @@ rm -f foo bar baz || framework_failure
 if cmp out.tar aout.tar > /dev/null ; then
 	printf "\nwarning: --diff test can't be run as root."
 else
-	"${TARLZ}" -q -df "${test3_lz}"
+	"${TARLZ}" -df "${test3_lz}" > /dev/null
 	[ $? = 1 ] || test_failed $LINENO
 	"${TARLZ}" -df "${test3_lz}" --ignore-ids || test_failed $LINENO
+	"${TARLZ}" -df "${test3_lz}" --exclude '*' || test_failed $LINENO
+	"${TARLZ}" -df "${in_tar_lz}" --exclude '*' || test_failed $LINENO
+	rm -f bar || framework_failure
+	"${TARLZ}" -df "${test3_lz}" foo baz --ignore-ids || test_failed $LINENO
+	"${TARLZ}" -df "${test3_lz}" --exclude bar --ignore-ids ||
+		test_failed $LINENO
+	rm -f foo baz || framework_failure
+	"${TARLZ}" -q -xf "${test3dir_lz}" || test_failed $LINENO
+	"${TARLZ}" -q -df "${test3dir_lz}" --ignore-ids || test_failed $LINENO
+	"${TARLZ}" -q -df "${test3dir_lz}" dir --ignore-ids || test_failed $LINENO
+	rm -rf dir || framework_failure
 fi
 rm -f out.tar aout.tar foo bar baz || framework_failure
 
@@ -505,10 +903,7 @@ rmdir dir1 || framework_failure
 rmdir dir1
 rm -f out.tar || framework_failure
 
-touch dummy_file || framework_failure
-if ln dummy_file dummy_link 2> /dev/null &&
-   ln -s dummy_file dummy_slink 2> /dev/null ; then
-	ln_works=yes
+if [ "${ln_works}" = yes ] ; then
 	name_100=name_100_bytes_long_nnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnn
 	path_100=dir1/dir2/dir3/path_100_bytes_long_nnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnn
 	path_106=dir1/dir2/dir3/path_longer_than_100_bytes_nnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnn
@@ -543,15 +938,12 @@ if ln dummy_file dummy_link 2> /dev/null &&
 	cmp "${in}" dir1/dir2/dir3/link || test_failed $LINENO
 	"${TARLZ}" -0 -q -c ../tmp/dir1 | "${TARLZ}" -x || test_failed $LINENO
 	diff -ru tmp/dir1 dir1 || test_failed $LINENO
-	rm -rf tmp/dir1 dir1 || framework_failure
+	rm -rf tmp dir1 || framework_failure
 	"${TARLZ}" -xf "${testdir}"/ts_in_link.tar.lz || test_failed $LINENO
 	"${TARLZ}" -df "${testdir}"/ts_in_link.tar.lz --ignore-ids ||
 		test_failed $LINENO
 	rm -f link1 link2 link3 link4 || framework_failure
-else
-	printf "\nwarning: skipping link test: 'ln' does not work on your system."
 fi
-rm -f dummy_slink dummy_link dummy_file || framework_failure
 
 printf "\ntesting long names..."
 
@@ -592,6 +984,8 @@ rm -f foo || framework_failure
 
 printf "\ntesting bad input..."
 
+mkdir dir1 || framework_failure
+cd dir1 || framework_failure
 "${TARLZ}" -q -xf "${testdir}"/dotdot1.tar.lz || test_failed $LINENO
 [ ! -e ../dir ] || test_failed $LINENO
 "${TARLZ}" -q -xf "${testdir}"/dotdot2.tar.lz || test_failed $LINENO
@@ -602,6 +996,8 @@ printf "\ntesting bad input..."
 [ ! -e dir ] || test_failed $LINENO
 "${TARLZ}" -q -xf "${testdir}"/dotdot5.tar.lz || test_failed $LINENO
 [ ! -e dir ] || test_failed $LINENO
+cd .. || framework_failure
+rm -rf dir1 || framework_failure
 
 dd if="${in_tar}" of=truncated.tar bs=1000 count=1 2> /dev/null
 "${TARLZ}" -q -tf truncated.tar > /dev/null
@@ -611,6 +1007,58 @@ dd if="${in_tar}" of=truncated.tar bs=1000 count=1 2> /dev/null
 [ ! -e test.txt ] || test_failed $LINENO
 rm -f truncated.tar || framework_failure
 
+# test --delete with split 'bar' tar member
+for i in 1 2 3 4 ; do
+	cat "${testdir}"/test3_sm${i}.tar.lz > out.tar.lz || framework_failure
+	for j in bar baz ; do
+		"${TARLZ}" -q -f out.tar.lz --delete $j
+		[ $? = 2 ] || test_failed $LINENO "$i $j"
+	done
+	cmp "${testdir}"/test3_sm${i}.tar.lz out.tar.lz || test_failed $LINENO $i
+	"${TARLZ}" -q -f out.tar.lz --delete foo
+	[ $? = 2 ] || test_failed $LINENO $i
+	"${TARLZ}" -xf out.tar.lz || test_failed $LINENO
+	[ ! -e foo ] || test_failed $LINENO
+	cmp cbar bar || test_failed $LINENO
+	cmp cbaz baz || test_failed $LINENO
+	rm -f out.tar.lz foo bar baz || framework_failure
+done
+
+# test format violations
+if [ "${ln_works}" = yes ] ; then
+	mkdir dir1 || framework_failure
+	"${TARLZ}" -C dir1 -xf "${t155}" || test_failed $LINENO
+fi
+for i in 1 2 3 ; do
+	"${TARLZ}" -q -tf "${testdir}"/t155_fv${i}.tar
+	[ $? = 2 ] || test_failed $LINENO $i
+	"${TARLZ}" -q -tf "${testdir}"/t155_fv${i}.tar --permissive ||
+		test_failed $LINENO $i
+	if [ "${ln_works}" = yes ] ; then
+		mkdir dir2 || framework_failure
+		"${TARLZ}" -C dir2 -xf "${testdir}"/t155_fv${i}.tar --permissive ||
+			test_failed $LINENO $i
+		diff -ru dir1 dir2 || test_failed $LINENO $i
+		rm -rf dir2 || framework_failure
+	fi
+done
+for i in 1 2 3 4 5 6 ; do
+	"${TARLZ}" -q -tf "${testdir}"/t155_fv${i}.tar.lz
+	[ $? = 2 ] || test_failed $LINENO $i
+	"${TARLZ}" -q -tf "${testdir}"/t155_fv${i}.tar.lz --permissive ||
+		test_failed $LINENO $i
+	if [ "${ln_works}" = yes ] ; then
+		mkdir dir2 || framework_failure
+		"${TARLZ}" -C dir2 -xf "${testdir}"/t155_fv${i}.tar.lz --permissive ||
+			test_failed $LINENO $i
+		diff -ru dir1 dir2 || test_failed $LINENO $i
+		rm -rf dir2 || framework_failure
+	fi
+done
+if [ "${ln_works}" = yes ] ; then
+	rm -rf dir1 || framework_failure
+fi
+
 # test compressed and --keep-damaged
 rm -f test.txt || framework_failure
 for i in "${inbad1}" "${inbad2}" ; do
